/* 
    added by Joerg Schulenburg 2006
    to switch between super computers (with slow disk and enough memory) 
    and small PCs (where we dont have enough memory) define USE_FILE
    to store and read the terms sequentially
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "Externals.h"
#include "FunctionPrototypes.h"
extern int myid;

//
// id: 0=gs_terms, 1=terms, 2=exp_terms 3=exc_terms
//
#ifdef USE_FILE
FILE *fp[4]={NULL,NULL,NULL,NULL};
#define  BUFSIZE (1024*1024l*8)
char buf[4][BUFSIZE];           // file cache
const char *name_ext[4]={"_gs",".terms",".exp_terms",".exc_terms"}
#else
// better have a growing list of memory blocks instead of realloc
//   realloc needs: Memory=oldmem+newmem and additional Memcopy, Jun08
#define ADDLEN (1024*1024l)     // num of terms per malloc block = 32MB
#define NUMBUFS     (1024l)     // maxMalloc = 32 * ADDLEN * NUMBUFS = 32GB
#define TermCompression 1
#if TermCompression                    // compress from 32B to 24B (-25%)
// MAX_NUM_CONFIGURATIONS.max=2^(32-8)=16e6 (see Constants.h)
#if (MAX_NUM_CONFIGURATIONS<<8)>>8 != MAX_NUM_CONFIGURATIONS
# error "no compression possible"
#endif
# define DESCSIZE (sizeof(int)*5)      // descriptors size (5*4 Byte)
# define MPLYSIZE (sizeof(float))      //  multiplier size   (4 Byte)
#else
# define DESCSIZE (sizeof(int)*6)      // descriptors size (6*4 Byte)
# define MPLYSIZE (sizeof(double))     //  multiplier size   (8 Byte)
#endif
#define TERMSIZE (DESCSIZE+MPLYSIZE)  //    one term size  (32 Byte)
// t_buf: list of storage buffer pointers, only first elements initialized
char   *t_buf[NUMBUFS+1][4];
int     t_blk[4]={-1,-1,-1,-1}; // index(*char) to buffer block
size_t  t_idx[4]={ADDLEN*TERMSIZE,
                  ADDLEN*TERMSIZE,
                  ADDLEN*TERMSIZE,
                  ADDLEN*TERMSIZE}; // pointer to next element [in block]
#endif

/* id: 0=gs_terms 1=terms 2=exp_terms 3=exc_terms */
#ifdef USE_FILE
int term_open(int id, char *opt){
  char name[256];
  strcpy(name,Control.name); strcat(name, name_ext[id]);
  if (myid==0) printf(" fopen(%s) - slow USE_FILE mode!\n",name);
  fp[id]=fopen(name,opt);
  if (!fp[id]) {
    printf("Cannot Open %s for %s...\nExiting!!\n\n", name, opt);
    ERR(1);
  }
  setvbuf(fp[id],buf[id],_IOFBF,BUFSIZE);
  return 0;
}
#endif

void term_rewind(int id) {
#ifdef USE_FILE
  if (!fp[id]) { term_open(id,"r"); }
  rewind(fp[id]);
#else
  // cause initialization to t_buf[0][id] of first access
  t_blk[id]=-1;
  t_idx[id]=TERMSIZE*ADDLEN;
#endif
}

/* set error if storing was not possible,
   skip further storing, break getting  */
int error=0;

/*
 *  gs_terms:  id=0
 *  terms:     id=1
 *  exp_terms: id=2 (ToDo) Jun08
 */
int term_store(int id, int *descriptors, double multiplier) {
#ifdef USE_FILE
  int nio;
  if (error) return 0; // skip on error
  if (!fp[id]) { term_open(id,"w+"); }
  nio =fwrite(descriptors,sizeof(   int),6,fp[id]);
  nio+=fwrite(&multiplier,sizeof(double),1,fp[id]);
  if (nio!=7) {
    error=1;
    printf("%3d: term_store fwrite_error=%d L%d\n",
                 myid, ferror(fp[id]), __LINE__);
    printf(" Terms can not be stored, you will get no results!\n"
           " Program will stop after all terms are counted.\n");
  }
#else
  if (error) return 0; // skip on error (dont store, but count)
  if   (t_idx[id] >= TERMSIZE*ADDLEN) {
    if (t_blk[id] == NUMBUFS-1) {
         printf("WARNING: not enough memory (NUMBUFS exceeded)\n"); }
    t_blk[id]++;        // extend the memory buffer
    t_idx[id]=0;
    // realloc has the problem of fragmentation and copying (max RAM/2)
    t_buf[t_blk[id]][id]=(char *)malloc(TERMSIZE*ADDLEN);
    if (!t_buf[t_blk[id]][id]) {
      printf("%3d: malloc failed, bkl= %10.2e [+=%.2e*%d]\n",
             myid, 1.*t_blk[id], 1.*ADDLEN, TERMSIZE);
      printf(" Further terms can not be stored, you will get no results!\n"
             " Program will stop after all terms are counted.\n");
      printf(" If you have fast storage, try USE_FILE which is slower\n");
      error=2;
      return -1;
    }
  }
#if TermCompression
  float m=multiplier;
  int d[5]; 
  d[0]=(descriptors[0]<<8)+descriptors[1];
  d[1]=descriptors[2];
  d[2]=descriptors[3];
  d[3]=descriptors[4];
  d[4]=descriptors[5];
  memcpy(t_buf[t_blk[id]][id]+t_idx[id],  d, DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(t_buf[t_blk[id]][id]+t_idx[id], &m, MPLYSIZE); t_idx[id]+=MPLYSIZE;
#else
  memcpy(t_buf[t_blk[id]][id]+t_idx[id], descriptors, DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(t_buf[t_blk[id]][id]+t_idx[id], &multiplier, MPLYSIZE); t_idx[id]+=MPLYSIZE;
#endif
#endif
  return 0;
}

           /* id: 0=gs_terms, 1=terms */
// returning 0 on OK, else exit
int term_get(int id, int *descriptors, double *multiplier) {
  if (error) {
     printf("Error term_get: Not all terms could be stored!\n"
            "Error term_get: Take the counters to estimate what you need\n");
     ERR(1);
     // ToDo: repeat slow calculation of a single element here?
  }
#ifdef USE_FILE   // bad for MPI Cluster
  int nio;
  if (!fp[id]) { term_open(id,"r"); }
  nio =fread(descriptors,sizeof(   int),6,fp[id]);
  nio+=fread(multiplier ,sizeof(double),1,fp[id]);
  if (nio!=7) {
    printf("%3d: term_get fread_error=%d L%d\n",
                 myid, ferror(fp[id]), __LINE__); ERR(1); }
#else
  if (t_idx[id] >= TERMSIZE*ADDLEN) {
    t_blk[id]++; t_idx[id]=0;
    if (t_blk[id] >= NUMBUFS) {
      printf("%3d: access beyond t_buf\n",myid); ERR(1); }
  }
#if TermCompression
  float m;
  int d[5];
  memcpy( d,t_buf[t_blk[id]][id]+t_idx[id], DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(&m,t_buf[t_blk[id]][id]+t_idx[id], MPLYSIZE); t_idx[id]+=MPLYSIZE;
  descriptors[0]=d[0]>>8;
  descriptors[1]=d[0]&255;
  descriptors[2]=d[1];
  descriptors[3]=d[2];
  descriptors[4]=d[3];
  descriptors[5]=d[4];
  *multiplier=(double)m;
#else
  memcpy(descriptors,t_buf[t_blk[id]][id]+t_idx[id], DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(multiplier ,t_buf[t_blk[id]][id]+t_idx[id], MPLYSIZE); t_idx[id]+=MPLYSIZE;
#endif
#endif
  return 0;
}

// Test:
//  g++ -g -DTEST -Wall storage.C
//  valgrind -v --tool=memcheck ./a.out
#ifdef TEST
int myid=0;
int main() {
  int i;
  int x[6]={0,1,2,3,4,5};
  double y=1.2345;
  for (i=0;i<10;i++) term_store(0,x,y);
  for (i=0;i<11;i++) term_get(0,x,&y);
}
#endif

// (Joerg Jun08: term_get3 replaced by id=3!)
